#include <pgpUtilities.h>
#include <pgpPFLErrors.h>
#include <pgpEncode.h>
#include <pgpCBC.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>

#include "support.h"
#include "interface.h"

#include "action.h"
#include "global.h"
#include "eventi.h"
#include "varie.h"

static PGPUInt32 GetKeySizeForAlgorithm(PGPCipherAlgorithm algorithm) {
  PGPUInt32 keySize = 0;
	
  switch (algorithm) {
    default:
      keySize = 0;
      break;
		
    case kPGPCipherAlgorithm_IDEA:
    case kPGPCipherAlgorithm_CAST5:
      keySize = 128 / 8;
      break;
		
    case kPGPCipherAlgorithm_3DES:
      keySize = 192 / 8;
      break;
  }
	
  return keySize;
}

void encode(const char *filename, PGPKeyID *signKeyID, gboolean detach, gboolean clear, PGPKeySetRef recKeySet) {
  PGPFileSpecRef inFileRef = kInvalidPGPFileSpecRef;
  PGPFileSpecRef outFileRef = kInvalidPGPFileSpecRef;
  PGPOptionListRef outList = kInvalidPGPOptionListRef;
  PGPKeyRef key;
  PGPError err = kPGPError_NoErr;
  char *outFileName, *pp = NULL;
  gint result;

  err = PGPNewFileSpecFromFullPath(pgpContext, filename, &inFileRef);
  if (IsntPGPError(err)) {
    outFileName = g_strconcat(filename, ".asc", NULL);
    err = PGPNewFileSpecFromFullPath(pgpContext, outFileName, &outFileRef);
    g_free(outFileName);
    if (IsntPGPError(err)) {

      PGPNewOptionList(pgpContext, &outList);

      if (signKeyID) {		// sign the file
        err = PGPGetKeyByKeyID(pgpKeySet, signKeyID, kPGPPublicKeyAlgorithm_Invalid, &key);
        if (IsntPGPError(err)) {
	  PGPKeySetRef tmpKeySet;

          GtkWidget *pass_dlg = create_passphrase();
          GtkWidget *lista = lookup_widget(pass_dlg, "clist1");

          PGPNewSingletonKeySet(key, &tmpKeySet);
          populate_lista(tmpKeySet, lista);
          PGPFreeKeySet(tmpKeySet);
            
	  result = gnome_dialog_run(GNOME_DIALOG(pass_dlg));
          switch (result) {
            case 0: {
              GtkWidget *pass_entry = lookup_widget(pass_dlg, "pass");
              pp = gtk_editable_get_chars(GTK_EDITABLE(pass_entry), 0, -1);

              gnome_dialog_close(GNOME_DIALOG(pass_dlg));
              break;
            } 
            case 2: {
              gnome_dialog_close(GNOME_DIALOG(pass_dlg));
              break;
            }
            default: {
              break;
            }
          }

          PGPAppendOptionList(outList,
            PGPOSignWithKey(pgpContext, key,
              PGPOPassphrase(pgpContext, pp),
              PGPOLastOption(pgpContext)),
            PGPOLastOption(pgpContext));
          g_free(pp);
  
          if (detach) {
            PGPAppendOptionList(outList,
              PGPODetachedSig(pgpContext,
	        PGPOLastOption(pgpContext)),
              PGPOLastOption(pgpContext));
          };
          if (clear) {
            PGPAppendOptionList(outList,
              PGPOClearSign(pgpContext, TRUE),
              PGPOLastOption(pgpContext));
          }
          g_free(pp);

        }
      }
      				// encrypt
      if (PGPKeySetRefIsValid(recKeySet)) {
        PGPAppendOptionList(outList,
          PGPOEncryptToKeySet(pgpContext, recKeySet),
          PGPOLastOption(pgpContext));
      }

      err = PGPEncode(pgpContext,
        PGPOFileNameString(pgpContext, filename),
        PGPOInputFile(pgpContext, inFileRef),
	PGPOOutputFile(pgpContext, outFileRef),
        PGPOArmorOutput(pgpContext, TRUE),
        PGPOSendNullEvents(pgpContext, 100),
        PGPOEventHandler(pgpContext, myEventHandler, NULL),
        PGPOAskUserForEntropy(pgpContext, TRUE),
        outList,
        PGPOLastOption(pgpContext));
    }
  }
  show_error(err);

  if (PGPFileSpecRefIsValid(inFileRef)) PGPFreeFileSpec(inFileRef);
  if (PGPFileSpecRefIsValid(outFileRef)) PGPFreeFileSpec(outFileRef);
}

static void decode(const char *filename) {
  PGPFileSpecRef inFileRef = kInvalidPGPFileSpecRef;
  PGPError err;

  err = PGPNewFileSpecFromFullPath(pgpContext, filename, &inFileRef);
  if (IsntPGPError(err)) {
    err = PGPDecode(pgpContext,
      PGPOInputFile(pgpContext, inFileRef),
      PGPOSendNullEvents(pgpContext, 100),
      PGPOKeySetRef(pgpContext, pgpKeySet),
      PGPOEventHandler(pgpContext, myEventHandler, NULL),
      PGPOSendEventIfKeyFound(pgpContext, TRUE),
		    //      PGPORecursivelyDecode(pgpContext, TRUE),
      PGPOLastOption(pgpContext));
  }
   
  show_error(err);

  if (PGPFileSpecRefIsValid(inFileRef)) PGPFreeFileSpec(inFileRef);
}

void open_filename_decode(GtkWidget *sender, gpointer user_data) {
  gchar *filename;
  GtkWidget *selector = (GtkWidget *)user_data;

  filename = gtk_file_selection_get_filename(GTK_FILE_SELECTION(selector));
  gtk_widget_hide(GTK_WIDGET(selector));
  decode(filename);
  gtk_widget_destroy(GTK_WIDGET(selector));
}

void hash_file(const char *filename, PGPHashAlgorithm algID) {
  PGPHashContextRef hashContext;
  PGPError err;
  PGPSize hashSize;
  GtkWidget *appBar, *dialog;
  gfloat completionPercent;
  void *buf;
  char *hash;
  
  int fd, letti, total = 0;
  long filesize;
  
  err = PGPNewHashContext(pgpMemoryMgr, algID, &hashContext);
  if (IsntPGPError(err)) {
    fd = open(filename, O_RDONLY);
    if (fd != -1) {
    	
      appBar = lookup_widget(app1, "appbar1");
      filesize = lseek(fd, 0L, SEEK_END);
      lseek(fd, 0L, SEEK_SET);
    	
      buf = g_malloc(16384);
      
      while ((letti = read(fd, buf, 16384)) > 0) {
      	total += letti;
        completionPercent = (1.0 * total) / filesize;

        gnome_appbar_set_progress(GNOME_APPBAR(appBar), completionPercent);
        gtk_main_iteration_do(FALSE);

        PGPContinueHash(hashContext, buf, letti);
      }
      close(fd);
      g_free(buf);
      PGPGetHashSize(hashContext, &hashSize);
      buf = g_malloc(hashSize);
      PGPFinalizeHash(hashContext, buf);
      hash = g_malloc(3 * hashSize + 1);
      fingerprint_string(buf, hashSize, hash);
      g_free(buf);

      dialog = gnome_message_box_new(hash,
        GNOME_MESSAGE_BOX_INFO,
	GNOME_STOCK_BUTTON_OK,
        NULL);
      gtk_window_set_modal(GTK_WINDOW(dialog), TRUE);
      gnome_dialog_run_and_close(GNOME_DIALOG(dialog));
    }
  }

  if (PGPHashContextRefIsValid(hashContext)) PGPFreeHashContext(hashContext);
}

void encrypt_decrypt_file(const char *filename, PGPCipherAlgorithm algID, char *key, gboolean encrypt) {
  PGPSymmetricCipherContextRef cipherContext = kInvalidPGPSymmetricCipherContextRef;
  PGPCBCContextRef chainingContext = kInvalidPGPCBCContextRef;
  PGPError err;
  PGPSize keySize, blockSize;

  GtkWidget *appBar;

  int fdin, fdout, letti, total = 0, keyLen, bigSize, effSize;
  long filesize;
  gfloat completionPercent;
  
  char *filename_out = NULL;
  
  char *bufin, *bufout;
  
  err = PGPNewSymmetricCipherContext(pgpMemoryMgr, algID, 
    GetKeySizeForAlgorithm(algID), &cipherContext);
  if (IsntPGPError(err)) {
    PGPGetSymmetricCipherSizes(cipherContext, &keySize, &blockSize);
    keyLen = strlen(key);
    if (keyLen < keySize) {
      key = g_realloc(key, keySize);
      memset(&key[keyLen], 0, keySize - keyLen);
    }
    err = PGPNewCBCContext(cipherContext, &chainingContext);
    cipherContext = kInvalidPGPSymmetricCipherContextRef;
    if (IsntPGPError(err)) {
      err = PGPInitCBC(chainingContext, key, NULL);
      if (IsntPGPError(err)) {
       fdin = open(filename, O_RDONLY);
        filename_out = g_strconcat(filename, ".CBC", NULL);
        fdout = open(filename_out, O_WRONLY | O_CREAT | O_TRUNC, S_IREAD | S_IWRITE);
        g_free(filename_out);
        if (fdin != -1 && fdout != -1) {
    	
          appBar = lookup_widget(app1, "appbar1");
          filesize = lseek(fdin, 0L, SEEK_END);
          lseek(fdin, 0L, SEEK_SET);

          bigSize = 16384 / blockSize;
          bigSize *= blockSize;

          bufin = g_malloc(bigSize);
          bufout = g_malloc(bigSize);      

          while ((letti = read(fdin, bufin, bigSize)) > 0) {
      	    total += letti;
            completionPercent = (1.0 * total) / filesize;

            gnome_appbar_set_progress(GNOME_APPBAR(appBar), completionPercent);
            gtk_main_iteration_do(FALSE);

            effSize = (letti - 1) / blockSize;
            effSize = (effSize + 1) * blockSize;
            if (effSize != letti) {
              memset(&bufin[letti], 0, effSize - letti);
            }

            if (encrypt) {
              err = PGPCBCEncrypt(chainingContext, bufin, effSize, bufout);
            } else {
              err = PGPCBCDecrypt(chainingContext, bufin, effSize, bufout);
            }
            
            write(fdout, bufout, effSize);
          }
          
          g_free(bufin);
          g_free(bufout);
        }
        if (fdin != -1) close(fdin);
        if (fdout != -1) close(fdout);
      }
    }
    g_free(key);
  }

  if (PGPCBCContextRefIsValid(chainingContext)) PGPFreeCBCContext(chainingContext);
  if (PGPSymmetricCipherContextRefIsValid(cipherContext)) PGPFreeSymmetricCipherContext(cipherContext);

  show_error(err);
}
